/*------------------------------------------------------------------------------
* correlator_e.c: correlator
*
* 2010/10/15 0.1 new
*-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>

#ifndef NOSSE2
#include <emmintrin.h>
#endif

#define PI      3.14159265358979323
#define LSBC    (1.0/16.0)          /* carrier LSB */
#define NC      32                  /* number of devided cycle */
#define CHIPR   1023000.0           /* code chip rate (chip/s) */

extern void gencode(int sat, double coff, double cs, int ns, short *code);

/* dot product ---------------------------------------------------------------*/
static int dot_int16(const short *a, const short *b, int n)
{
#ifdef NOSSE2
    const short *p=a,*q=b,*r=a+n;
    int sum=0;
    while (p<r) sum+=(*p++)*(*q++);
    return sum;
#else
    const short *p=a,*q=b,*r=a+n;
    int sum[4];
    __m128i x,y,z;
    
    z=_mm_setzero_si128();
    
    for (;p<r;p+=8,q+=8) {
        x=_mm_loadu_si128((__m128i *)p);
        y=_mm_loadu_si128((__m128i *)q);
        x=_mm_madd_epi16(x,y);
        z=_mm_add_epi32(z,x);
    }
    _mm_storeu_si128((void *)sum,z);
    
    return sum[0]+sum[1]+sum[2]+sum[3];
#endif
}
/* multiply carrier ----------------------------------------------------------*/
static void mulcarr(double freq, double phi0, double ts, int ns,
                    const char *buff, short *I, short *Q)
{
    static short cost[NC+1]={0},sint[NC+1]={0};
    double circ;
    const char *p,*q=buff+ns;
    int i,index;
    
    /* generate table */
    if (cost[0]==0.0) {
        for (i=0;i<NC+1;i++) {
            cost[i]=(short)(cos(2.0*PI/NC*i)/LSBC);
            sint[i]=(short)(sin(2.0*PI/NC*i)/LSBC);
        }
    }
    /* to devided cycle */
    ts  *=freq*NC;
    phi0*=NC/2.0/PI;
    
    circ=phi0-ts;
    circ-=floor(circ/NC)*NC; /* 0<=circ<NC */
    
    for (p=buff;p<q;p++) {
        circ+=ts;
        index=((int)circ)&0x1F; /* max: 2^31/32/38.2MHz=1.757s */
        *I++=cost[index]*(*p);
        *Q++=sint[index]*(*p);
    }
    for (;p<q+8;p++) { /* zero padding */
        *I++=*Q++=0;
    }
}
/* correlator ----------------------------------------------------------------*/
extern void correlator(int sat, double coff, double freq, double phi0,
                       const char *buff, double ts, int ns, int s, int nc,
                       double *I, double *Q)
{
    short *Is,*Qs,*code;
    int i,j,k,na=1+2*s*nc;
    
    if (!(Is  =(short *)malloc(sizeof(short)*(ns+8)))||
        !(Qs  =(short *)malloc(sizeof(short)*(ns+8)))||
        !(code=(short *)malloc(sizeof(short)*(ns+na+8)))) return;
    
    /* multiply carrier */
    mulcarr(freq,phi0,ts,ns,buff,Is,Qs);
    
    /* generate code */
    gencode(sat,coff,ts*CHIPR,ns,code+s*nc);
    
    for (i=0,j=ns+s*nc;i<s*nc;i++,j++) {
        code[i]=code[i+ns];
        code[j]=code[i+s*nc];
    }
    /* zero padding */
    for (i=ns+na;i<ns+na+8;i++) code[i]=0;
    
    /* multiply code and sum */
    I[0]=dot_int16(Is,code+s*nc,ns)*LSBC; /* I_P */
    Q[0]=dot_int16(Qs,code+s*nc,ns)*LSBC; /* Q_P */
    
    for (k=1;k<=nc;k++) {
        I[2*k-1]=dot_int16(Is,code+s*(nc-k),ns)*LSBC; /* I_E */
        Q[2*k-1]=dot_int16(Qs,code+s*(nc-k),ns)*LSBC; /* Q_E */
        I[2*k  ]=dot_int16(Is,code+s*(nc+k),ns)*LSBC; /* I_L */
        Q[2*k  ]=dot_int16(Qs,code+s*(nc+k),ns)*LSBC; /* Q_L */
    }
    free(Is); free(Qs); free(code);
}
